/* Author: Samuel Hornus <samuel.hornus@inria.fr>
 * Copyright © Inria, 2015
 * Licence: Creative Commons CC BY-ND 3.0 license available online at
 * http://creativecommons.org/licenses/by-nd/3.0/
 */

#include "spherical.h"
#include <iostream>

using namespace std;

bool SphericalPolygon::closest_pt_on_tet(Vec3f & dir) {
  return false;
}

bool SphericalPolygon::add_and_nearest(const Vec3f & silVertex, Vec3f & dir) {
  // return true if the updated simplex contains the origin
  switch( size() ) {
    case 0: {
              emplace_back(silVertex);
              dir = begin()->sil_;
              return false;
              break;
            }
    case 1: {
              Vec3f clipNorth = silVertex.normalized();
              float dot = begin()->north_ | clipNorth;
              if( dot < -0.99984769515 ) { // about one degree
                // intersection of two almost opposite hemispheres ==> empty
                clear();
                return true;
              } else if( dot > 0.99984769515 ) {
                dir = clipNorth;
                return false;
              }
              Vec3f v(Vec3f::cross(clipNorth, begin()->north_).LInfNormalized());
              begin()->vertex_ = v;
              emplace_back(v.negated(), clipNorth, silVertex);
              Vec3f AB = begin()->sil_ - silVertex;
              dir = (silVertex - (((silVertex | AB) / AB.squaredLength()) * AB)).LInfNormalized();
              return false;
              break;
            }
    case 2: {
              Vec3f clipNorth = silVertex.LInfNormalized();
              iterator next = begin();
              iterator cur = next++;
              float vDot = begin()->vertex_ | clipNorth;
              if( vDot >= _ray_spherical_eps ) {
                // we'll get a triangle
                next->vertex_ = Vec3f::cross(clipNorth, next->north_).LInfNormalized();
                Vec3f v(Vec3f::cross(cur->north_, clipNorth).LInfNormalized());
                emplace(next, v, clipNorth, silVertex);
                // Now, find the new direction to probe
                // Order of the vertices is: old0, new, old1

                Vec3f AB = data()[0].sil_ - silVertex;
                Vec3f AC = data()[2].sil_ - silVertex;
                Vec3f N = Vec3f::cross(AB, AC); // normal, pointing toward the origin
                if( (N|silVertex) >= 0.0f ) cerr << "Bad orientation of normal in Spherical [A]";

                if( (N|data()[0].vertex_) > 0.0f ) {
                  cerr << "Bad configuration in Spherical [B]\n";
                }
                if( (N|data()[1].vertex_) > 0.0f ) {
                  // The closest point to the origin lies on the edge sil[0]--sil[1]
                  dir = Vec3f::cross(data()[1].vertex_, AB).LInfNormalized();
                  return false;
                }
                if( (N|data()[2].vertex_) > 0.0f ) {
                  // The closest point to the origin lies on the edge sil[1]--sil[2]
                  dir = Vec3f::cross(AC, data()[2].vertex_).LInfNormalized();
                  return false;
                }
                // The closest point to the origin lies in the triangle
                dir = N.LInfNormalized();
                return false;

              } else if( vDot <= - _ray_spherical_eps ) {
                // we'll get a triangle
                cur->vertex_ = Vec3f::cross(clipNorth, cur->north_).LInfNormalized();
                Vec3f v(Vec3f::cross(next->north_, clipNorth).LInfNormalized());
                emplace_back(v, clipNorth, silVertex);
                // Now, find the new direction to probe
                // Order of the vertices is: old0, old1, new

                Vec3f AB = data()[0].sil_ - silVertex;
                Vec3f AC = data()[1].sil_ - silVertex;
                Vec3f N = Vec3f::cross(AC, AB); // normal, pointing toward the origin
                if( (N|silVertex) >= 0.0f ) cerr << "Bad orientation of normal in Spherical [A]";

                if( (N|data()[1].vertex_) > 0.0f ) {
                  cerr << "Bad configuration in Spherical [C]\n";
                }
                if( (N|data()[2].vertex_) > 0.0f ) {
                  // The closest point to the origin lies on the edge sil[1]--sil[2]
                  dir = Vec3f::cross(data()[2].vertex_, AC).LInfNormalized();
                  return false;
                }
                if( (N|data()[0].vertex_) > 0.0f ) {
                  // The closest point to the origin lies on the edge sil[2]--sil[0]
                  dir = Vec3f::cross(AB, data()[0].vertex_).LInfNormalized();
                  return false;
                }
                // The closest point to the origin lies in the triangle
                dir = N.LInfNormalized();
                return false;

              } else {
                // we keep a moon crescent
                float curTest(clipNorth | Vec3f::cross(cur->north_, cur->vertex_));
                Vec3f nextTest(Vec3f::cross(next->north_, next->vertex_));
                if( curTest > 0.0f ) {
                  if( (clipNorth | nextTest) <= 0.0f ) {
                    next->north_ = clipNorth;
                    next->sil_   = silVertex;
                    cur->vertex_ = Vec3f::cross(next->north_, cur->north_);
                    cur->vertex_.LInfNormalize();
                    next->vertex_ = cur->vertex_;
                    next->vertex_.negate();
              Vec3f AB = cur->sil_ - silVertex;
              dir = (silVertex - (((silVertex | AB) / AB.squaredLength()) * AB)).LInfNormalized();
              return false;
                  } else {
                    // the crescent is unchanged
                    std::cerr << "kept a crescent\n";
                  }
                } else {
                  if( (clipNorth | nextTest) > 0.0f ) {
                    cur->north_   = clipNorth;
                    cur->sil_     = silVertex;
                    next->vertex_ = Vec3f::cross(cur->north_, next->north_);
                    next->vertex_.LInfNormalize();
                    cur->vertex_ = next->vertex_;
                    cur->vertex_.negate();
              Vec3f AB = next->sil_ - silVertex;
              dir = (silVertex - (((silVertex | AB) / AB.squaredLength()) * AB)).LInfNormalized();
              return false;
                  } else {
                    //std::cerr << "killed a crescent\n";
                    clear();
                    return true;
                  }
                }
              }
              break;
            }
            /*
    case 3: {
              pts_[3] = silVertex;
              return closest_pt_on_tet(dir);
              break;
            }
            */
    default: std::cerr << "The IMPOSSIBLE just happened in GJK::nearest"; break;
  }
  return false;
}
